
/* -*-mode: C; style: K&R; c-basic-offset: 4 ; -*- */

/*
 *  silver.c - Main code.
 *
 *  Silver - A simple webcam grabbing application
 *
 * Homepage:
 *   http://www.gnump3d.org
 *
 * Author:
 *  Steve Kemp <skx@tardis.ed.ac.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  Steve Kemp
 *  ---
 *  http://www.steve.org.uk/
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ghttp.h>
#include <unistd.h>

#include "silver.h"



/*
 * Some default settings.
 */
#ifndef TMP_IMAGE_NAME
# define TMP_IMAGE_NAME "/tmp/img.jpg"
#endif /* TMP_IMAGE_NAME */

#ifndef DEFAULT_CAM
# define DEFAULT_CAM "http://localhost/shot.jpg"
#endif /* DEFAULT_CAM */


#ifndef MAX_LINE_LENGTH
# define MAX_LINE_LENGTH 1024
#endif /* MAX_LINE_LENGTH */




/*
 * Global variables.
 */
GtkWidget *window;    /* Main window */
GtkWidget *image;     /* Image widget */
GtkTooltips *toolTip;   /* Displayed over the image. */
GtkWidget *url_entry; /* Text input field. */
GtkWidget *dir_entry; /* Text input field. */


/*
 * The URL we're gonna open, the default save directory, and
 * the current count of captured images.
 */
char url[256];
char dir[256];
static int saved = 0;



/**
 * Download the current webcam image, and update the image
 * object to display it.
 */
gint get_image(gpointer data)
{
    ghttp_request *req  = NULL;
    ghttp_status status;
    int code = 0;

    gdk_threads_enter();
    req = ghttp_request_new();

    if ( ghttp_set_uri( req, url ) != 0 )
    {
	g_message("Failed - wrong url %s\n", url );
	gdk_threads_leave() ;
	return FALSE;
    }

    if(ghttp_prepare(req) != 0) 
    {
	if(req)
	{
	    ghttp_request_destroy(req);
	}

	g_message("Unable to prepare request.\n");
	gdk_threads_leave() ;
	return FALSE;
    }


    status = ghttp_process( req );
    code   = ghttp_status_code(req);
    
    if ( code == 200 )
    {
	int length  = ghttp_get_body_len(req);
	int written = 0;
	char *body  = ghttp_get_body(req);

	FILE *f = fopen( TMP_IMAGE_NAME, "wb" );
	if ( f == NULL )
	{
		g_message("FAILED TO OPEN\n" );
		perror( "Arse");
		return FALSE;
	}

	written = fwrite(body, (sizeof(char)),length, f);

	if ( written != length )
	    {
		g_message("FAILED TO WRITE ALL OF FILE\n" );
		printf("Wrote %d\n", written );
		return FALSE;
	    }
	fclose( f );
	{
	    /* Update the image. */
	    gtk_image_set_from_file(GTK_IMAGE(image), TMP_IMAGE_NAME );
	}
	
    }

    ghttp_request_destroy( req );

    /* Update the tooltip */
    gtk_tooltips_set_tip( toolTip,
			  window,
			  url,
			  0 );

    gdk_threads_leave() ;
    return( TRUE);
}


/*
 * Save the currently displayed image to our save directory.
 */
gint save_image(gpointer data)
{
    char cmdbuff[256];
    memset( cmdbuff, sizeof(cmdbuff), '\0' );

    gdk_threads_enter();

    snprintf( cmdbuff, (sizeof(cmdbuff)-1), 
	      "/bin/cp " TMP_IMAGE_NAME " %s/capture-%d.jpg",
	      dir, saved );

    /* Debug our saving command line */
    g_message( cmdbuff );
    
    system( cmdbuff );
    saved += 1;

    gdk_threads_leave();
    return 0;
}



/**
 * Force a reload of the image in response to the mouse.
 */
static void file_reload( GtkWidget *w, gpointer   data )
{
    get_image( 0 );
}


/*
 * Update the URL we're currently viewing.  This is called
 * by our menu items.
 */
static void update_location( GtkWidget *w, gpointer data )
{
    memset(url,sizeof(url), '\0');

    strncpy(url, (char*)data, sizeof(url)-1 );

    /* Force reload */
    get_image(0);
}


static void half_size( GtkWidget *w, gpointer data )
{
    if ( gtk_check_menu_item_get_active( data ) )
	{
	    int x, y;

	    /* get size */
	    gtk_window_get_size( GTK_WINDOW(window), &x, &y );

	    printf("Current %d,%d\n", x, y );

	    x /= 2;
	    y /= 2;

	    /* Make sure it did get resized. */
	    gtk_window_get_size( GTK_WINDOW(window), &x, &y );
	    printf("New %d,%d\n", x, y );
	}
}
/*
 *  Toggle the use of tooltips - called via our menu item.
 */
static void toggle_tooltips( GtkWidget *w, gpointer   data )
{
    if ( gtk_check_menu_item_get_active(data) )
    {
	if ( toolTip != NULL )
	{
	    gtk_tooltips_enable(toolTip);
	}
	else
	{
	    g_message("toolTip == NULL at %s:%d\n",
		      __FILE__, __LINE__ );
	}
    }
    else
    {
	if ( toolTip != NULL )
	{
	    gtk_tooltips_disable(toolTip);
	}
	else
	{
	    g_message("toolTip == NULL at %s:%d\n",
		      __FILE__, __LINE__ );
	}
    }
}

/**
 * Display a dialog box.
 */
void quick_message(gchar *message) 
{

   GtkWidget *dialog, *label, *okay_button;
   
   /* Create the widgets */
   
   dialog = gtk_dialog_new();
   label = gtk_label_new (message);
   okay_button = gtk_button_new_with_label("Okay");

   /* Ensure that the dialog box is destroyed when the user clicks ok. */
   gtk_signal_connect_object (GTK_OBJECT (okay_button), "clicked",
                              GTK_SIGNAL_FUNC (gtk_widget_destroy), dialog);
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->action_area),
                      okay_button);

   /* This makes it so the button is the default. */
   GTK_WIDGET_SET_FLAGS (okay_button, GTK_CAN_DEFAULT);
   gtk_widget_grab_default (okay_button);


   /* Add the label, and show everything we've added to the dialog. */
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
                      label);
   
   gtk_window_position ((GtkWindow*)dialog, GTK_WIN_POS_CENTER);
   gtk_widget_show_all (dialog);
}



/**
 * Display the about dialog box.
 */
static void help_about( GtkWidget *w, gpointer   data )
{
  quick_message( "\n"
		 "Silver - v" VERSION
		 "\n\n"
		 "A simple webcam grabbing application\n"
		 "\n"
		 "By Steve Kemp"
		 "\n"
		 "http://www.steve.org.uk"
		 "\n");
}


/**
 * Settings dialog box is killed..
 */
static void setup_end( GtkWidget *w, gpointer data )
{
    /* Get the text. */
    const gchar *u = gtk_entry_get_text( GTK_ENTRY(url_entry) );
    const gchar *d = gtk_entry_get_text( GTK_ENTRY(dir_entry) );

    g_message("New URL: %s\n", u );
    memset(url,sizeof(url), '\0');
    strncpy( url, u,sizeof(url)-1 );

    g_message("New dir: %s\n", d );
    memset( dir,sizeof(dir), '\0');
    strncpy( dir, d, sizeof(dir)-1 );

    gtk_widget_destroy(w);

    /* Force reload */
    get_image(0);
}


/**
 * Display our settings.
 */
void do_setup() 
{
   GtkWidget *dialog, *okay_button;
   GtkWidget *url_label, *dir_label;
   
   /* Create the widgets */
   dialog    = gtk_dialog_new();
   url_label = gtk_label_new ("URL:");
   dir_label = gtk_label_new ("Save Dir:");

   /* Create the entry field - and make sure it's got the current url in it */
   url_entry = gtk_entry_new ();
   gtk_entry_set_max_length ( GTK_ENTRY(url_entry), 100);
   gtk_entry_set_text( GTK_ENTRY(url_entry), url );

   /* Save directory. */
   dir_entry = gtk_entry_new ();
   gtk_entry_set_max_length ( GTK_ENTRY(dir_entry), 100);
   gtk_entry_set_text( GTK_ENTRY(dir_entry), dir );

   /* OK button */
   okay_button = gtk_button_new_with_label("Okay");


   /* Ensure that the dialog box is destroyed when the user clicks ok. */
   
   gtk_signal_connect_object (GTK_OBJECT (okay_button), "clicked",
                              GTK_SIGNAL_FUNC (setup_end), dialog);
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->action_area),
                      okay_button);

   /* Add the label, and show everything we've added to the dialog. */
   {
       GtkWidget *urlbox, *dirbox;

       urlbox = gtk_hbox_new (FALSE, 0);
       dirbox = gtk_hbox_new (FALSE, 0);

       gtk_box_pack_start (GTK_BOX (urlbox), url_label, TRUE, TRUE, 0);
       gtk_box_pack_end (GTK_BOX (urlbox), url_entry, TRUE, TRUE, 0);

       gtk_box_pack_start (GTK_BOX (dirbox), dir_label, TRUE, TRUE, 0);
       gtk_box_pack_end (GTK_BOX (dirbox), dir_entry, TRUE, TRUE, 0);

       gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
			  urlbox);
       gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
			  dirbox);
       
   }

   /* CEnter the window and display it. */
   gtk_window_position ((GtkWindow*)dialog, GTK_WIN_POS_CENTER);
   gtk_widget_show_all (dialog);

}


GtkWidget* setup_menubar( GtkWidget *window )
{
  /*-- Declare the GTK Widgets used in the program --*/
  GtkWidget *menuFile;
  GtkWidget *menuView;  
  GtkWidget *menubar;
  GtkWidget *menu;
  GtkWidget *menuitem;


  /*-- Create the menu bar --*/
  menubar = gtk_menu_bar_new();

  
  /*---------------- Create File menu items ------------------*/

  menuFile = gtk_menu_item_new_with_label ("File");
  gtk_menu_bar_append (GTK_MENU_BAR(menubar), menuFile);
  gtk_widget_show(menuFile);

  /*-- Create File submenu  --*/
  menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuFile), menu);
  gtk_widget_show (menuitem);


  /*-- Create Setup menu item under File submenu --*/
  menuitem = gtk_menu_item_new_with_label ("Setup");
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_signal_connect(GTK_OBJECT (menuitem), "activate", 
		     GTK_SIGNAL_FUNC (do_setup), NULL);
  gtk_widget_show (menuitem);

  /*-- Seperator --*/
  menuitem = gtk_separator_menu_item_new();
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show (menuitem);

  /*-- Create Exit menu item under File submenu --*/
  menuitem = gtk_menu_item_new_with_label ("Exit");
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_signal_connect(GTK_OBJECT (menuitem), "activate", 
		     GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
  gtk_widget_show (menuitem);


  /*-- Create View submenu --*/
  menu = gtk_menu_new();

  /*---------------- Create View menu items --------------------*/
  menuView = gtk_menu_item_new_with_label ("View");
  gtk_menu_bar_append (GTK_MENU_BAR(menubar), menuView);
  gtk_widget_show(menuView);

  /*-- Create View submenu --*/
  menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuView), menu);

  /*-- Create Refresh menu item under File submenu --*/
  menuitem = gtk_menu_item_new_with_label ("Refresh");
  gtk_signal_connect(GTK_OBJECT (menuitem), "activate", 
		     GTK_SIGNAL_FUNC (file_reload), NULL);
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show( menuitem );

  /*-- Create Refresh menu item under File submenu --*/
  menuitem = gtk_check_menu_item_new_with_label ("Half-size");
  gtk_signal_connect(GTK_OBJECT (menuitem), "activate", 
  		     GTK_SIGNAL_FUNC (half_size), GTK_OBJECT(menuitem));
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show( menuitem );

  /*-- Create Refresh menu item under File submenu --*/
  menuitem = gtk_check_menu_item_new_with_label ("Tooltips");
  gtk_signal_connect(GTK_OBJECT (menuitem), "activate", 
  		     GTK_SIGNAL_FUNC (toggle_tooltips), GTK_OBJECT(menuitem));
  /* Default to enabled. */
  gtk_check_menu_item_set_active ( GTK_CHECK_MENU_ITEM(menuitem), 1 );
  gtk_menu_append(GTK_MENU(menu), menuitem);

  gtk_widget_show( menuitem );




  /*-- Create View submenu --*/
  menu = gtk_menu_new();

  /*---------------- Create Sites menu items --------------------*/
  menuView = gtk_menu_item_new_with_label ("Sites");
  gtk_menu_bar_append (GTK_MENU_BAR(menubar), menuView);
  gtk_widget_show(menuView);

  /*-- Create Sites submenu --*/
  menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuView), menu);

  /*-- Create tearoff menu item under Sites submenu --*/
  menuitem = gtk_tearoff_menu_item_new();
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_widget_show (menuitem);

  /* Load the websites */
  {
      /* ~/.silverrc */
      int readFile = 0;
      FILE *f;
      char *home = getenv( "HOME" );
      
      if ( home != NULL )
      {
	  char *tmp  = malloc( strlen( home ) + strlen( "/.silverrc" ) + 2 );
	  if ( tmp != NULL )
	  {
	      strcpy( tmp, home );
	      strcat( tmp, "/.silverrc" );
	      f = fopen( tmp, "r" );
	      if ( f != NULL )
	      {
		  fclose( f );
		  load_sites(menu, tmp);
		  readFile = 1;
		  free( tmp );
	      }
	  }
      }

      if ( readFile == 0 )
      {
	  /* /etc/silverrc */
	  f = fopen( "/etc/silverrc", "r" );
	  if ( f != NULL )
	  {
	      fclose( f );
	      load_sites(menu, "/etc/silverrc" );
	  }
      }
  }


  /* Help menu */
  menuFile = gtk_menu_item_new_with_label ("Help");
  gtk_menu_item_right_justify( GTK_MENU_ITEM(menuFile) );
  gtk_menu_bar_append (GTK_MENU_BAR(menubar), menuFile);
  gtk_widget_show(menuFile);

  /*-- Create Help submenu  --*/
  menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuFile), menu);
  gtk_widget_show (menuitem);

  /*-- Create About menu item under Help submenu --*/
  menuitem = gtk_menu_item_new_with_label ("About");
  gtk_menu_append(GTK_MENU(menu), menuitem);
  gtk_signal_connect(GTK_OBJECT (menuitem), "activate", 
		     GTK_SIGNAL_FUNC (help_about), NULL);
  gtk_widget_show (menuitem);

  return( menubar);
}


/**
 * Handle mouse clicks on the applications main window.
 */
static gint button_press_event( GtkWidget      *widget,
                                GdkEventButton *event )
{
  if (event->button == 1)
  {
      /* left mouse button -> Reload image. */
      get_image(0);
  }
  else if (event->button == 3)
  {
      /* Right mouse button -> Save current image. */
      save_image(0);
  }

  return TRUE;
}



/**
 * Initialize the main window.
 */
int main( int   argc, char *argv[] )
{
    GtkWidget *vbox;
  
    GtkWidget *menubar;

    /* Don't allow root to run this program - Gtk programs aren't safe.. */
    if ( 0 == geteuid() )
    {
	fprintf( stderr, "Root shouldn't play this game.\n" );
	exit( -1 );
    }
 
    /* Initialize */
    gtk_init (&argc, &argv);

    memset(url,sizeof(url), '\0');
    memset(dir,sizeof(dir), '\0');

    if ( argc > 1 )
    {
	strncpy( url, argv[1], sizeof(url)-1 );
    }
    else
    {
	/* Default image. */
	strcpy(url, DEFAULT_CAM );
    }

    /* Default save location. */
    strcpy( dir, "/tmp" );

    /* Create main window */
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_name (window, "Silver");


    /* Packing box */
    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), vbox);
    gtk_widget_show (vbox);
    
    /* Make sure we can exit cleanly */
    g_signal_connect (G_OBJECT (window), "destroy",
		      G_CALLBACK (gtk_main_quit), NULL);
    

    /* Create and install menu bar */
    menubar = setup_menubar( window);
    gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, TRUE, 0);

    /* Create the image and dislay it */
    image = gtk_image_new_from_file( TMP_IMAGE_NAME );
    gtk_box_pack_start (GTK_BOX (vbox), image, TRUE, TRUE, 0);


    /* Create the tooltip object. */
    toolTip = gtk_tooltips_new();
    gtk_tooltips_set_tip( toolTip,
			  window,
			  url,
			  0 );

    gtk_tooltips_enable(toolTip);
 


    /* Event signals */
    gtk_widget_set_events (window, GDK_BUTTON_PRESS_MASK );
    g_signal_connect (G_OBJECT (window), "button_press_event",
		      G_CALLBACK (button_press_event), NULL);

    /* Reload the image every three seconds unless otherwise
     * notified.
     */
    gtk_timeout_add(3000, get_image, window);

    /* Show the widgets */
    gtk_window_position ((GtkWindow*)window, GTK_WIN_POS_CENTER);
    gtk_widget_show_all (window);

    /* Run the event loop. */
    gtk_main ();
    
    return 0;
}


/*
 *  read a line from a file.
 */
int
readLine( FILE *inFile, char *line ) 
{
    int i = 0;
    int charIn;

    while ( (charIn = fgetc(inFile)) != EOF ) 
    {
	line[i++] = charIn;
	if ( charIn == '\n' )
	    break;
	if (i >= MAX_LINE_LENGTH - 1) 
	{
	    fprintf( stderr, "warning, line longer than buffer: truncated\n");
	    break;
	}
    }
    if ( (i == 0) && (charIn == EOF) ) 
    {
	return(-1);
    } 
    else
    {
	line[i] = '\0';  /* terminate string */
    }
    return(i);
} 


/*
 * Load the sites that are available.
 *
 * Read the file sites.txt and insert each valid line that we find.
 * into the 
 */
int load_sites(GtkWidget *menu, char *fileName)
{
    char buff[MAX_LINE_LENGTH];
    FILE  *f     = NULL;
    int lineLen  = 0;
    int lineCount= 0;

    printf( "Reading config file: %s\n", fileName );
    f = fopen( fileName, "r" );
    if ( f == NULL )
    {
	return -1;
    }
    
    while ( (lineLen = readLine( f, buff )) >= 0 )
    {
	/* Skip comments. */
	if ( buff[0] != '#' )
	{
	    char label[128];
	    char url[256];
	    char *equals;
	    char *c_url;   /* Copy of url */
	    char *c_label; /* Copy of label */

	    memset( label, sizeof(label),'\0');
	    memset( url, sizeof(url),'\0');

	    /* Parse line. */
	    equals = strstr( buff, "=" );
	    if ( equals != NULL )
	    {
		GtkWidget *menuitem ;
		int labelLen = equals-buff;

		strncpy( label, buff, labelLen);
		label[labelLen]= '\0';
			
		strcpy( url, equals+1);
		url[strlen(url)-1]= '\0';

		printf("Label '%s' - URL '%s'\n",
		       label, url );
		
		/* Copy the URL. */
		c_url = (char *)malloc( strlen( url ) + 1 );
		strcpy( c_url, url );
		
		/* Copy the label. */
		c_label = (char *)malloc( strlen( label ) + 1 );
		strcpy( c_label, label );

		/* Create the new menu item. */
		menuitem = gtk_menu_item_new_with_label (c_label);
		gtk_menu_append(GTK_MENU(menu), menuitem);
		gtk_signal_connect(GTK_OBJECT (menuitem), "activate", 
				   GTK_SIGNAL_FUNC (update_location), c_url);
		gtk_widget_show (menuitem);


		lineCount += 1;
	    }
	}
    }
    fclose(f);

    return( lineCount );
}
